#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : fneg.py
# Contents: determine false negative rates for napa
# Author  : Christian Borgelt
# History : 2014.02.06 file created from detect.py (cocoexp)
#           2014.03.25 extended to cover influence region tests
#           2014.03.25 extended to cover time binning tests
#           2014.03.28 adapted to time bins statistics in napa
#           2014.07.03 smaller sizes added for point-centered bins
#-----------------------------------------------------------------------
from sys        import argv, path
from os.path    import join
from time       import strftime
from random     import seed
from os         import remove, devnull
from subprocess import check_output
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par     import get_args, project, run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts       import genpsts

#-----------------------------------------------------------------------

oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------

def napa (fname, width=0.003, mode='w', stat='cpc', alpha=1,
          surr=1000, seed=0):
    cmd = ['napa', '-i1', '-w%g' % width,
           '-m'+mode, '-s'+stat, '-x%d' % alpha,
           '-n%d' % surr, '-S%d' % seed, fname, '-']
    with open(devnull, 'w') as null:
        for line in check_output(cmd,stderr=null).split('\n')[:-1]:
            line = line.split() # run program and get the p-value
            if line[0] != 'item': res = line[1]; break
    return float(res) if res[-1] in '0123456789' else float(res[:-1])

#-----------------------------------------------------------------------

def fneg (ncnt=100, rate=20, period=3,      # basic     parameters
          size=5, coin=5, jitter=0.003,     # assembly  parameters
          width=0.003, surr=1000,           # detection parameters
          mode='wc', stat='cpc', alpha=1,   # detection parameters
          runs=1000, verbose=100, srng=1):  # other     parameters
    if jitter < 0: jitter = width
    srng += 1023*size           # init. seed and get data file name
    name  = 'n%d_p%g_r%g_s%d_c%d_j%g_w%g_%s_%s%d_s%d' \
          % (ncnt,period,rate,size,coin,jitter*1000,
             width*1000,mode,stat,alpha,surr)
    data  = join('data', name+'.txt')
    asm   = set(range(1,size+1))# create the assembly to inject
    pvals = []                  # initialize the result list
    for i in range(runs):       # repeat the experiment multiple times
        if verbose > 0 and i % verbose == 0:
            print(strftime('%Y.%m.%d %H:%M:%S fneg ')+name+' '+str(i))
        seed(srng)              # seed random number generator
        psts = genpsts(ncnt, period, rate, jitter=jitter,
                       asms=[[asm,-coin]])
        with open(data, 'w') as out: # write the spike trains
            for k in range(len(psts)):
                for s in psts[k]: out.write('%d %.16g\n' % (k+1,s))
        pvals.append(napa(data, width, mode, stat, alpha, surr, srng))
        srng += coin            # execute the statistical test
    if verbose > 0:             # print a log message
        print(strftime('%Y.%m.%d %H:%M:%S fneg ')+name+' '+str(runs))
    remove(data)                # call napa/nass and collect p-values
    with open(join('data','fn_'+name+'.txt'), 'w') as out:
        for p in pvals: out.write('%.16g\n' % p)
    return pvals                # store and return p-values

#-----------------------------------------------------------------------

def collect (ncnt=100, rate=20, period=3,        # basic     parameters
             sizes=[5], coins=[5], jitter=0.003, # assembly  parameters
             width=0.003, surr=1000,             # detection parameters
             mode='w', stat='cpc', alpha=1,      # detection parameters
             siglvl=0.01):                       # other     parameters
    if jitter < 0: jitter = width
    res = []                    # initialize the result
    for s in sizes:             # traverse the sizes
        for c in coins:         # traverse the coincidences
            name = 'n%d_p%g_r%g_s%d_c%d_j%g_w%g_%s_%s%d_s%d' \
                 % (ncnt,period,rate,s,c,jitter*1000,
                    width*1000,mode,stat,alpha,surr)
            n = k = 0           # init. the counters
            with open(join('data','fn_'+name+'.txt'), 'r') as inp:
                for line in inp:# traverse the p-values
                    if float(line) > siglvl: k += 1
                    n += 1      # count the false negatives
            if k > 0: res.append((s,c,k/float(n)))
    name = 'n%d_p%g_r%g_j%g_w%g_%s_%s%d_s%d' \
         % (ncnt,period,rate,jitter*1000,
            width*1000,mode,stat,alpha,surr)
    with open(join('data','fn_'+name+'.txt'), 'w') as out:
        for x in res: out.write('%d %d %.16g\n' % x)

#-----------------------------------------------------------------------

if __name__ == '__main__':
    grids = [ [ ( 'runs',   [ 1000 ] ),
                ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( 'jitter', [ 0.000] ), # exact coincidences
                ( 'width',  [ 0.003] ),
                ( 'surr',   [ 1000 ] ),
                ( 'mode',   [ 'b' ] ),
                ( 'stat',   [ 'cpc', 'cxc', 'ccf', 'ccr',
                              'cif', 'cir', 'ciw', 'cwr', 'cpo' ] ),
                ( 'alpha',  [ 1, 2, 3, 4 ] ),
                ( 'size',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ),
                ( 'coin',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ) ],
              [ ( 'runs',   [ 1000 ] ),
                ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( 'jitter', [ 0.003] ), # jittered coincidences
                ( 'width',  [ 0.003] ),
                ( 'surr',   [ 1000 ] ),
                ( 'mode',   [ 'b', 'p', 'w', 'i' ] ),
                ( 'stat',   [ 'cpc', 'cxc', 'ccf', 'ccr',
                              'cif', 'cir', 'ciw', 'cwr' ] ),
                ( 'alpha',  [ 1, 2, 3, 4 ] ),
                ( 'size',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ),
                ( 'coin',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ) ],
              [ ( 'runs',   [ 1000 ] ),
                ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( 'jitter', [ 0.003] ), # jittered coincidences
                ( 'width',  [ 0.003] ),
                ( 'surr',   [ 1000 ] ),
                ( 'mode',   [ 'b', 'p' ] ),
                ( 'stat',   [ 'cpo' ] ),
                ( 'alpha',  [ 1, 2, 3, 4 ] ),
                ( 'size',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ),
                ( 'coin',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ) ],
              [ ( 'runs',   [ 1000 ] ),
                ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( 'jitter', [ 0.003] ), # smaller windows
                ( 'width',  [ 0.0015, 0.002, 0.0025 ] ),
                ( 'surr',   [ 1000 ] ),
                ( 'mode',   [ 'p' ] ),
                ( 'stat',   [ 'cpc', 'cxc', 'ccf', 'ccr',
                              'cif', 'cir', 'ciw', 'cwr', 'cpo' ] ),
                ( 'alpha',  [ 1, 2, 3, 4 ] ),
                ( 'size',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ),
                ( 'coin',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ) ] ]

#    grids = [ [ ( 'runs',   [ 1000 ] ),
#                ( 'ncnt',   [  100 ] ),
#                ( 'period', [    3 ] ),
#                ( 'rate',   [   20 ] ),
#                ( 'jitter', [ 0.003] ),
#                ( 'width',  [ 0.0015, 0.002, 0.0025 ] ),
#                ( 'surr',   [ 1000 ] ),
#                ( 'mode',   [ 'p' ] ),
#                ( 'stat',   [ 'cif', 'cir', 'ciw', 'cwr', 'cpo' ] ),
#                ( 'alpha',  [ 1, 2, 3, 4 ] ),
#                ( 'size',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ),
#                ( 'coin',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3 ] ) ] ]

    if len(argv) < 2 or argv[1] != 'collect':
        run(fneg, grids, int(argv[1])-1 if len(argv) > 1 else -1)
    else:
        siglvl = float(argv[2]) if len(argv) > 2 else 0.01
        ranges = dict(project(grids, ['size','coin'])[0])
        grids  = project(grids, ['ncnt','period','rate','jitter',
                                 'width','mode','stat','alpha','surr'])
        for i in range(get_args(grids)):
            collect(sizes=ranges['size'], coins=ranges['coin'],
                    siglvl=siglvl, **get_args(grids, i))
